#include <assert.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/io.h>
#include <sys/dir.h>

#include "usb-xhci.h"

#define TRB_STRUCT_SIZE 16
#define CMDRING_TRB_NUM 256
#define MAXSLOTS 64

uint32_t mmio_addr = 0xfebf0000;
uint32_t mmio_size = 0x4000;

char *xhci_mmio_base;
int fd;
#define PAGE_SHIFT      12
#define PAGE_SIZE       (1 << PAGE_SHIFT)
#define PFN_PRESENT     (1ull << 63)
#define PFN_PFN         ((1ull << 55) - 1)
#define PHY_RAM         0x80000000


struct XHCITRB *cmd_ring_trbs;
struct XHCIEvRingSeg *seg;
struct Parameter *address_parameter;
struct Parameter *configure_parameter;

uint64_t *dcbaap;

struct xhci_op_regs *xhci_op_regs;
struct xhci_run_regs *xhci_run_regs;
struct xhci_db_regs *xhci_db_regs;


void* mem_map( const char* dev, size_t offset, size_t size )
{
    int fd = open( dev, O_RDWR | O_SYNC );
    if ( fd == -1 ) {
        return 0;
    }

    void* result = mmap( NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, offset );

    if ( !result ) {
        return 0;
    }

    close( fd );
    return result;
}

uint32_t page_offset(uint32_t addr)
{
    return addr & ((1 << PAGE_SHIFT) - 1);
}
uint64_t gva_to_gfn(void *addr)
{
	uint64_t pme, gfn;
	size_t offset;
	offset = ((uintptr_t)addr >> 9) & ~7;
	lseek(fd, offset, SEEK_SET);
	read(fd, &pme, 8);
	if (!(pme & PFN_PRESENT))
		return -1;
	gfn = pme & PFN_PFN;
	return gfn;
}

uint64_t gva_to_gpa(void *addr)
{
	uint64_t gfn = gva_to_gfn(addr);
	assert(gfn != -1);
	return (gfn << PAGE_SHIFT) | page_offset((uint64_t)addr);
}


uint32_t gen_control_field(uint32_t trb_type, uint32_t slotid, uint32_t epid) {
    uint32_t control = rand();

    // set type
    control &= 0xFFFF03FF;        // clear trb_type
    control |= (trb_type << 10);  // set trb_type

    // set slot id
    control &= 0x00FFFFFF;
    control |= (slotid << 24);

    // set epid
    control &= 0xffe0ffff;
    control |= (epid << 16);

    control |= 0x1;             // set cycle bit to be 1 to equal to ccs.
    return control;
}

static uint64_t cmd_ring_trbs_builder(void) {
    
    int i, bufsize = TRB_STRUCT_SIZE * CMDRING_TRB_NUM;

    cmd_ring_trbs = (struct XHCITRB *)aligned_alloc(PAGE_SIZE, bufsize);
    for (i = 0; i < MAXSLOTS; i++) {
        cmd_ring_trbs[i].control = gen_control_field(9, 0, 0);
    }
}

static uint64_t parameters_builder(void) {
    
    int i, bufsize = 32 * 33;

    address_parameter = (uint64_t *)malloc(bufsize);
    configure_parameter = (uint64_t *)malloc(bufsize);

}

static uint64_t evring_seg_builder(void) {
    seg = (struct XHCIEvRingSeg *)malloc(sizeof(struct XHCIEvRingSeg));
    seg->addr_low = 0x1000;
    seg->addr_high = 0;
    seg->size = 1000;

}

uint64_t *dev_ctx_builder(void) {
  uint64_t *dev_ctx_pointer;
  dev_ctx_pointer = (uint64_t *)malloc(32 * 32);

  return dev_ctx_pointer;
}

static uint64_t dev_ctx_array_builder(void) {
  // build device context base address arrary
  int i;
  dcbaap = (uint64_t *)malloc((MAXSLOTS+1) * 8);

  for (i = 1; i < MAXSLOTS; i++) {
      dcbaap[i] = gva_to_gpa(dev_ctx_builder());
  }
}

static void xhci_kick_ep(uint32_t slotid, uint32_t epid){
    
    uint64_t reg, val;

    reg = slotid << 2;
    val &= 0xFF00;        
    val |= epid;              

    xhci_db_regs->db[reg] = val;
}


void process(){
    struct Parameter parameter;
    uint32_t control;
    uint32_t slotid = 5;
    uint32_t epid = 5;
    uint32_t portid = 1;

    control = gen_control_field(CR_ADDRESS_DEVICE, slotid, epid);
    control &= ~(1<<9);
    parameter.ictl_ctx[0] = 0;
    parameter.ictl_ctx[1] = 3;

    parameter.slot_ctx[0] = 0;
    parameter.slot_ctx[1] &= 0xff00ffff;
    parameter.slot_ctx[1] |= (portid << 16);
    cmd_ring_trbs[0].control = control;
    memcpy(address_parameter, &parameter, sizeof(parameter));
    cmd_ring_trbs[0].parameter = gva_to_gpa(address_parameter);


    control = gen_control_field(CR_CONFIGURE_ENDPOINT, slotid, epid);
    control &= ~(1<<9);

    parameter.ictl_ctx[0] &= 0xfffffffc;
    parameter.ictl_ctx[1] &= 0xfffffffd;
    parameter.ictl_ctx[1] |= 0x1;

    parameter.ictl_ctx[0] &= ~(1 << epid);
    parameter.ictl_ctx[1] |= 1 << epid;


    cmd_ring_trbs[1].control = control;
    memcpy(configure_parameter, &parameter, sizeof(parameter));
    cmd_ring_trbs[1].parameter = gva_to_gpa(configure_parameter);
    xhci_op_regs->crcr = gva_to_gpa(cmd_ring_trbs);   // send ring segment
    xhci_db_regs->db[0] = 0x0;            // process commands
    xhci_kick_ep(slotid, epid);
}



int main(int argc, char *argv[]){
    

    xhci_mmio_base = mem_map( "/dev/mem", mmio_addr, mmio_size );
    if ( !xhci_mmio_base ) {
        return 0;
    }
    fd = open("/proc/self/pagemap", O_RDONLY);
	if (fd < 0) {
		perror("open");
		exit(1);
	}

    xhci_op_regs = (struct xhci_op_regs *)(xhci_mmio_base + OFF_OPER);
    xhci_run_regs = (struct xhci_run_regs *)(xhci_mmio_base + OFF_RUNTIME);
    xhci_db_regs = (struct xhci_db_regs *)(xhci_mmio_base + OFF_DOORBELL);
    
    printf("MMIO base address = 0x%lx\n", xhci_mmio_base);
    cmd_ring_trbs_builder();
    evring_seg_builder();
    parameters_builder();
    dev_ctx_array_builder();

    xhci_op_regs->usbcmd = 0x2;   // reset all the xhci state
    xhci_op_regs->crcr = gva_to_gpa(cmd_ring_trbs); // set ring segment
    xhci_op_regs->dcbaap = gva_to_gpa(dcbaap);        // set dcbaap address
    xhci_run_regs->irs[0].erstsz = 1;     // enable event ring segment
    xhci_run_regs->irs[0].erstba = gva_to_gpa(seg);  // set event ring address
    xhci_run_regs->irs[0].erdp = 0x2000;  // set TRB address

    xhci_op_regs->usbcmd = 0x1; 
    xhci_db_regs->db[0] = 0x0;    // process commands

    process();
}

